from __future__ import annotations

import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from importlib import resources
from pathlib import Path
from typing import Annotated, Literal

import copier
import copier._vcs as copier_vcs
import typer
from ruamel.yaml import YAML, CommentedMap

from coasti.logger import log, setup_logging
from coasti.prompt import prompt_like_copier_from_template, prompt_single

from .product import ProductsConfig

yaml = YAML()
app = typer.Typer()


@app.command()
def list():
    """List installed products."""

    config = ProductsConfig()

    typer.echo(typer.style("products:", fg=typer.colors.BRIGHT_BLACK))
    for p in config.products:
        typer.echo(
            typer.style(f"- name: {p['id']}", fg=typer.colors.WHITE, bold=True)
        )
        for key, value in p.items():
            if key != "id":
                typer.echo(
                    typer.style(f"  {key}: {value}", fg=typer.colors.BRIGHT_BLACK)
                )


@app.command()
def add(
    vcs_repo: Annotated[
        str | None,
        typer.Argument(
            help="Url of the product's git repo.",
        ),
    ] = None,
    quiet: Annotated[
        bool,
        typer.Option(
            "-q",
            "--quiet",
            help="Don't ask questions, use all defaults, and overwrite.",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option(
            "-v",
            "--verbose",
            help="Enable debug logging.",
        ),
    ] = False,
):
    """Add a product to coasti."""

    if verbose:
        setup_logging("DEBUG")

    config = ProductsConfig()

    with resources.path("coasti", "") as module_path:
        p_res = prompt_like_copier_from_template(
            src_path=str(module_path / "product" / "questions"),
            data={"vcs_repo": vcs_repo},
            defaults=quiet,
        )
        pid = p_res.answers["id"]

    if pid in config.product_ids:
        if not quiet and not prompt_single(
            f"Product id {pid} already exists. Overwrite?", type=bool, default=True
        ):
            log.info("Exiting")
            raise typer.Exit(code=1)

    config.upsert_product_from_answers(p_res)
    config.save_products_config()

    log.info(f"Updated {pid} in {str(config.products_yaml_path)}")

    if prompt_single(f"Do you want to install {pid} now?", type=bool, default=True):
        install(pid)


@app.command()
def install(
    pid: Annotated[
        str | None,
        typer.Argument(
            help="Id of the product.",
        ),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option(
            "-v",
            "--verbose",
            help="Enable debug logging.",
        ),
    ] = False,
):
    """
    Fetch resources for a product that has already been added.

    Uses copier, git and details from config/products.yml
    """
    _install_or_update("install", pid, verbose)

@app.command()
def update(
    pid: Annotated[
        str | None,
        typer.Argument(
            help="Id of the product.",
        ),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option(
            "-v",
            "--verbose",
            help="Enable debug logging.",
        ),
    ] = False,
):
    """
    Update an installed product.

    Uses copier, git and details from config/products.yml
    """
    _install_or_update("update", pid, verbose)

def _install_or_update(
    method : Literal["update", "install"], pid: str | None, verbose: bool = True
):
    if verbose:
        setup_logging("DEBUG")

    config = ProductsConfig()
    if pid is None:
        pid = prompt_single(
            "Select the product to install:", type=str, choices=config.product_ids
        )

    if pid not in config.product_ids:
        log.error(
            f"{pid} not found in products. Available for install:\n"
            f"  {config.product_ids}"
        )
        raise typer.Exit(code=1)

    try:
        product = config.get_product(pid)
        if method == "install":
            product.install()
        elif method == "update":
            product.update()
        else:
            raise NotImplementedError
    except copier.ProcessExecutionError as e:
        log.error(f"Failed to {method} {pid}. Check your connection and authentication.")
        log.info(e)
        raise typer.Exit(code=1)
    except Exception as e:
        # use typer to exit and avoid stack trace (which might contain auth info).
        log.error(e)
        raise typer.Exit(code=1)



import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from importlib import resources
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Iterable, Mapping

from ruamel.yaml import YAML, CommentedMap

yaml = YAML()

if TYPE_CHECKING:
    from coasti.product.product import ProductDetails

class CoastiConfig:

    def __init__(self) -> None:
        self.products_config = None

    @property
    def base_dir(self):
        """Current projects coasti base directory."""
        coasti_root = Path(os.getenv("COASTI_BASE_DIR", Path.cwd()))
        return coasti_root



_config: CoastiConfig = None  # type: ignore


def get_config():
    global _config
    if _config is None:
        _config = CoastiConfig()
    return _config



import logging
from typing import Any

import typer

# Define a new log level for success (between INFO and WARNING)
SUCCESS_LEVEL_NUM = 25
logging.addLevelName(SUCCESS_LEVEL_NUM, "SUCCESS")


class TyperLogHandler(logging.Handler):
    # use logging with typer
    # https://github.com/fastapi/typer/issues/203#issuecomment-840690307

    def emit(self, record: logging.LogRecord) -> None:
        fg = None
        bg = None
        bold = False
        if record.levelno == logging.DEBUG:
            fg = typer.colors.BRIGHT_BLACK
        elif record.levelno == logging.INFO:
            fg = None
        elif record.levelno == SUCCESS_LEVEL_NUM:
            fg = typer.colors.GREEN
        elif record.levelno == logging.WARNING:
            fg = typer.colors.YELLOW
        elif record.levelno == logging.CRITICAL:
            fg = typer.colors.BRIGHT_RED
        elif record.levelno == logging.ERROR:
            fg = typer.colors.RED
            bold = True
        typer.secho(self.format(record), bg=bg, fg=fg, bold=bold)


# Create a type alias for our logger with extra attribute
class EnhancedLogger(logging.Logger):
    def success(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(SUCCESS_LEVEL_NUM):
            self._log(SUCCESS_LEVEL_NUM, msg, args, **kwargs)



def setup_logging(
    level: str | int = "INFO",
    propagate: bool = False,
    name: str = "coasti_installer",
) -> EnhancedLogger:
    """
    Configure our singleton logger instance with TyperLogHandler configured.

    Can be called in functions to update the module level log instance:

    ```python
    from .logger import log, setup_logging

    def test():
        log.info("Foo")
        setup_logging(level="DEBUG")
        log.debug("Bar")
    ```
    """
    global log

    # If we have an existing logger instance, clean it up
    if log is None:
        log = EnhancedLogger(name)
    elif log.name != name:
        log.name = name

    # Remove any existing handlers to avoid duplicates
    for handler in log.handlers[:]:
        log.removeHandler(handler)
        handler.close()

    # do not (!) set level on the log instance, or we will not be able to override it.
    log.propagate = propagate

    # Add our custom handler, define level here.
    typer_handler = TyperLogHandler()
    typer_handler.setLevel(level)
    typer_handler.setFormatter(logging.Formatter("%(message)s"))
    log.addHandler(typer_handler)

    return log


# Singleton logger instance, with default level
log = None
log = setup_logging()



from __future__ import annotations

import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from importlib import resources
from pathlib import Path
from typing import Literal, NotRequired, TypedDict

import copier
import copier._vcs as copier_vcs
from ruamel.yaml import YAML, CommentedMap

from coasti.logger import log
from coasti.prompt import PromptResponse

yaml = YAML()


class ProductsConfig:
    coasti_base_dir: Path
    products_config: CommentedMap

    def __init__(self, coast_base_dir: Path | None = None) -> None:
        self.coasti_base_dir = coast_base_dir or Path(
            os.getenv("COASTI_BASE_DIR", Path.cwd())
        )
        self._load_products_config()

    @property
    def products_yaml_path(self):
        """Get the abs path to the products yaml."""
        products_yaml_path = self.coasti_base_dir / "config" / "products.yml"
        if not products_yaml_path.is_file():
            raise ValueError(
                "Could not find config/products.yml. Call from a coasti project, "
                "or set COASTI_BASE_DIR"
            )
        return products_yaml_path

    @property
    def products(self) -> list[ProductDetails]:
        """Loaded product details from config/products.yml"""
        return self.products_config["products"]

    @property
    def product_ids(self) -> list[str]:
        return [p["id"] for p in self.products_config["products"]]

    def get_product_details(self, pid):
        return [p for p in self.products if p.get("id") == pid][0]

    def get_product(self, pid : str):
        details = self.get_product_details(pid)
        return Product(details=details, coasti_base_dir=self.coasti_base_dir)

    def _load_products_config(self):
        config = yaml.load(self.products_yaml_path)
        if not isinstance(config, CommentedMap) or "products" not in config.keys():
            raise ValueError(
                "Could not find the products section in config/products.yml."
            )
        # this is a list in yaml, but it might not have any entries, and be none.
        config["products"] = config.get("products") or []
        self.products_config = config
        return config

    def save_products_config(self):
        with self.products_yaml_path.open("w") as f:
            yaml.dump(self.products_config, f)

    def upsert_product_from_answers(self, p_res: PromptResponse[ProductDetails]):
        """
        Insert or update a product to our yaml based on a questionaire.
        The `.answers` in the PromptResponse should be ProductDetails.
        """

        p = Product(details=p_res.answers, coasti_base_dir=self.coasti_base_dir)
        p.save_secrets()

        pid = p_res.answers["id"]
        if pid in self.product_ids:
            product = [p for p in self.products if p["id"] == pid][0]
            product.update(p_res.answers_to_remember)
        else:
            self.products.append(p_res.answers_to_remember)


class Product:
    details: ProductDetails
    coasti_base_dir: Path

    def __init__(self, coasti_base_dir: Path, details: ProductDetails) -> None:
        self.details = details
        self.coasti_base_dir = coasti_base_dir

    @property
    def id(self):
        return self.details["id"]

    @property
    def secret_path(self):
        return self.coasti_base_dir / "config" / "secrets" / f"vcs_auth_{self.id}"

    @property
    def dst_path(self):
        return Path(self.details["dst_path"])

    @property
    def vcs_auth_token(self):
        if self.details["vcs_auth_type"] != "Auth Token":
            return None

        res = self.details.get("vcs_auth_token", None)
        if res is None:
            if not self.secret_path.exists():
                log.warning(f"{self.id} auth token neither in details nor in file")
            res = self.secret_path.read_text()
        return res

    @property
    def vcs_auth_sshkeypath(self):
        if self.details["vcs_auth_type"] != "SSH Key":
            return None

        res = self.details.get("vcs_auth_sshkeypath", None)
        if res is None:
            if not self.secret_path.exists():
                log.warning(f"{self.id} sshkeypath neither in details nor in file")
            res = self.secret_path.read_text()
        return res

    def save_secrets(self):
        """Take unmasked answers and save auth token or ssh key path to file."""
        if self.details["vcs_auth_type"] == "Auth Token":
            auth = self.details.get("vcs_auth_token")
        elif self.details["vcs_auth_type"] == "SSH Key":
            auth = self.details.get("vcs_auth_sshkeypath")
        else:
            log.debug(
                f"No secrets to write for auth type {self.details['vcs_auth_type']}"
            )
            return

        if auth is None:
            raise ValueError("To save secrets, provide them in ProductDetails.")

        self.secret_path.write_text(auth)

    def install(self):
        """
        Install this product by getting its resources via copier.
        Authentication is retrieved from disk and injected into the git commands.
        """

        # Clone template
        with copier_git_injection(
            https_token=self.vcs_auth_token,
            ssh_key_path=self.vcs_auth_sshkeypath,
        ):

            log.info(f"Using copier to install {self.id}. Downloading...")
            copier.run_copy(
                src_path=self.details["vcs_repo"],
                dst_path=self.dst_path,
                vcs_ref=self.details["vcs_ref"],
                unsafe=True,
            )

        self._create_symlinks()

    def update(self):
        """
        Update this product by getting its resources via copier.
        Authentication is retrieved from disk and injected into the git commands.
        """

        # Clone template
        with copier_git_injection(
            https_token=self.vcs_auth_token,
            ssh_key_path=self.vcs_auth_sshkeypath,
        ):
            log.info(f"Using copier to update {self.id}. Downloading...")
            copier.run_update(
                dst_path=self.dst_path,
                answers_file="config/install_answers.yml",
                unsafe=True,
            )

    def _create_symlinks(self):
        log.info(f"Creating symlinks for {self.id}")
        for part in ["config", "config/secrets", "data", "logs"]:
            if (dst := self.dst_path / part).exists():
                src = self.coasti_base_dir / part / self.id
                try:
                    src.symlink_to(dst)
                    log.debug(f"Created symlink {str(src)} -> {str(dst)}")
                except FileExistsError:
                    log.debug(f"Link target already exists ({str(dst)})")


class ProductDetails(TypedDict):
    # see questions/copier.yml
    vcs_repo: str
    id: str
    dst_path: str
    vcs_ref: str
    vcs_auth_type: Literal["skip", "Auth Token", "SSH Key"]
    vcs_auth_token: NotRequired[str]
    vcs_auth_sshkeypath: NotRequired[str]


@contextmanager
def copier_git_injection(
    *,
    https_token: str | None = None,
    ssh_key_path: str | Path | None = None,
) -> Iterator[None]:
    """
    Inject auth settings into all git commands executed by Copier.

    - https_token: used for HTTPS clones/fetches via GIT_ASKPASS.
    - ssh_key_path: absolute path to an SSH private key to force for SSH clones/fetches.

    We monkeypatch copiers get_git() command to keep env vars in a small scope.
    """
    if ssh_key_path is not None and https_token is not None:
        raise ValueError("Provide either https_token or ssh_key_path")

    if ssh_key_path is not None and not Path(ssh_key_path).is_absolute():
        raise ValueError("ssh_key_path must be an absolute path")

    original_get_git = copier_vcs.get_git

    try:
        extra_env = {}

        if https_token:
            with resources.as_file(
                resources.files("coasti.product").joinpath(
                    "askpass" + (".bat" if sys.platform == "win32" else ".sh")
                )
            ) as askpass_script:
                extra_env["GIT_ASKPASS"] = askpass_script
                # scripts simply return the token env var

            extra_env["GIT_AUTH_TOKEN"] = https_token
            extra_env["GIT_TERMINAL_PROMPT"] = "0"
            # Some git flows require this to force askpass in non-tty contexts:

        elif ssh_key_path:
            extra_env["GIT_SSH_COMMAND"] = (
                f"ssh -i {ssh_key_path} -o IdentitiesOnly=yes"
            )

        def patched_get_git():
            git = original_get_git()
            # Attach env to the command object.
            # (Plumbum supports cmd.with_env(VAR=...))
            cmd = git.with_env(**extra_env) if extra_env else git
            return cmd

        copier_vcs.get_git = patched_get_git
        yield
    finally:
        copier_vcs.get_git = original_get_git



import typer

from .init import app as init_app
from .product import app as product_app

app = typer.Typer()


app.add_typer(init_app) # only one command so far
app.add_typer(product_app, name="product", help="List, add or update products")
